import os
import sys
import pytest

ALL = set("darwin linux win32".split())


def pytest_runtest_setup(item):
    supported_platforms = ALL.intersection(mark.name for mark in item.iter_markers())
    plat = sys.platform
    if "darwin" in plat:
        plat = "linux"
    if supported_platforms and plat not in supported_platforms:
        pytest.skip("cannot run on platform %s" % (plat))


@pytest.fixture()
def no_fs(monkeypatch):
    def mockreturn(path):
        return False

    monkeypatch.setattr(os.path, "exists", mockreturn)



def make_string_pep440_compatible(raw_str):
    """
    pep440 only allows a subset of characters in the
    version name:

    - alphanumeric
    - dots

    this will restrict the string to that set.
    """
    final_chars = []
    for r in raw_str:
        if "a" <= r <= "z" or "A" <= r <= "Z" or "0" <= r <= "9" or r == ".":
            final_chars.append(r)
    return "".join(final_chars)



import os
import re
import logging
from packaging.version import parse
from .scm.git import Git
from .scm.base import DEFAULT_TAG_VERSION
from .exception import VersionerError
from .version_string import make_string_pep440_compatible

SCM_TYPES = [Git]
LOG = logging.getLogger(__name__)
RELEASE_FORMAT = "{main_version}"
FORMAT = "{main_version}.dev{commit_count}+{branch}.{scm_change_id}"


def get_version(
    path=os.curdir,
    is_release=False,
    version_format=FORMAT,
    release_version_format=RELEASE_FORMAT,
    version_file="VERSION",
    release_branch_regex=None,
    scm_type=None,
):
    """
    return the version.
    """
    path = os.path.abspath(path)
    version_file_path = os.path.join(path, version_file)

    scm = _get_scm(scm_type, path)
    if not scm:
        existing_version = _read_version_file(version_file_path)
        if existing_version:
            return existing_version

        if scm_type is None:
            msg = "unable to detect scm type."
        else:
            msg = "scm type {0} not found, or is not a valid repo.".format(
                scm_type
            )
        raise VersionerError(msg)

    version = determine_version(
        scm,
        version_format=version_format,
        release_version_format=release_version_format,
        is_release=is_release,
    )

    _write_version_file(version_file_path, version)
    return version


def determine_version(
    scm,
    version_format=FORMAT,
    release_version_format=RELEASE_FORMAT,
    release_branch_regex=None,
    is_release=False,
):
    props = scm.get_properties()

    release_branch_regex = release_branch_regex or scm.RELEASE_BRANCH_REGEX
    if not re.match(release_branch_regex, props["branch"]):
        LOG.info(
            "branch {0} does not match regex {1}. Using default tag version.".format(
                props["branch"], release_branch_regex
            )
        )
        props["main_version"] = DEFAULT_TAG_VERSION
    else:
        props["main_version"] = props["tag_version"]

    props["branch"] = make_string_pep440_compatible(props["branch"])

    fmt_to_use = release_version_format if is_release else version_format

    try:
        return str(parse(fmt_to_use.format(**props)))
    except KeyError as ke:
        raise VersionerError(
            "key {0} was not provided by the scm type {1}".format(
                ke, scm.get_name()
            )
        )


def _read_version_file(version_file):
    if not os.path.exists(version_file):
        return
    with open(version_file) as fh:
        return fh.read()


def _get_scm(scm_type, path):
    for SCMType in SCM_TYPES:
        if scm_type is None or scm_type == SCMType.get_name():
            if SCMType.is_repo(path):
                return SCMType(path)
    return None


def _write_version_file(version_file, version):
    with open(version_file, "w+") as fh:
        fh.write(version)



import pytest
from vcver.scm.mock import MockSCM
from vcver.version import determine_version


@pytest.fixture
def scm_dict():
    return {
        "tag_version": "1.0",
        "commit_count": 20,
        "scm_change_id": "abc123",
        "branch": "master",
    }


@pytest.fixture
def scm(scm_dict):
    return MockSCM(scm_dict)


@pytest.mark.parametrize(
    "inp, expected",
    [
        ({}, "1.0.dev20+master.abc123"),
        ({"branch": "develop"}, "0.dev20+develop.abc123"),
        ({"branch": "hotfix"}, "0.dev20+hotfix.abc123"),
        ({"branch": "feature/ABC-123"}, "0.dev20+featureabc123.abc123"),
        ({"commit_count": 30}, "1.0.dev30+master.abc123"),
        ({"tag_version": "2.0b1"}, "2.0b1.dev20+master.abc123"),
        ({"tag_version": "2.0beta"}, "2.0b0.dev20+master.abc123"),
        ({"scm_change_id": "012345"}, "1.0.dev20+master.12345"),
    ],
)
def test_version(inp, expected, scm_dict):
    scm_dict.update(inp)
    scm = MockSCM(scm_dict)
    assert determine_version(scm) == expected


def test_custom_version_format(scm):
    assert (
        determine_version(scm, version_format="{main_version}.{commit_count}")
        == "1.0.20"
    )


def test_release_version_format(scm):

    assert (
        determine_version(
            scm,
            release_version_format="{main_version}.{commit_count}",
            is_release=False,
        )
        == "1.0.dev20+master.abc123"
    )

    assert (
        determine_version(
            scm, is_release=True, release_version_format="{main_version}.{commit_count}"
        )
        == "1.0.20"
    )


def test_is_release(scm):
    assert determine_version(scm, is_release=False) == "1.0.dev20+master.abc123"

    assert determine_version(scm, is_release=True) == "1.0"



DEFAULT_TAG_VERSION = "0"


class SCM(object):
    """
    SCMs are wrappers around a particular form of version control.
    """

    RELEASE_BRANCH_REGEX = None

    REQUIRED_PARAMS = ["tag_version", "commit_count", "scm_change_id", "branch"]

    def __init__(self, path):
        self._path = path

    @classmethod
    def get_name(cls):
        return str(cls.__name__.lower())

    @staticmethod
    def is_repo(path):
        """
        should return True if it is a valid
        repo of the provided scm type.
        """
        return False

    def get_properties(self):
        """
        return back a dictionary containing
        all properties exposed by this scm.
        """
        return {}


def extract_tag_version(tag):
    """
    extract a version string from a tag, or None if
    it is not a valid numerical version.
    """
    if tag.startswith("v"):
        return tag[1:]
    return None



from .base import SCM


class MockSCM(SCM):
    """ a mock scm """

    RELEASE_BRANCH_REGEX = "master"

    def __init__(self, props):
        self._props = props

    @staticmethod
    def is_repo(path):
        return True

    def get_properties(self):
        return self._props



import os
import logging
import subprocess
from .base import SCM, extract_tag_version, DEFAULT_TAG_VERSION
from ..exception import VersionerError

CMD_LATEST_VERSION_TAG = "git describe --tags --match 'v*' --abbrev=0"
CMD_BRANCH = "git rev-parse --abbrev-ref HEAD"
CMD_SHOW_REFS = 'git branch --points-at=HEAD --format="%(refname)"'
CMD_REV_LIST = "git rev-list HEAD --reverse"
CMD_LIST_TAGS = "git tag --list"
CMD_REV_COUNT = "git rev-list {start}..{end} --count"
CMD_SHORT_HASH = "git rev-parse --short HEAD"

LOG = logging.getLogger(__name__)


class GitCommandError(VersionerError):
    pass


class Git(SCM):

    RELEASE_BRANCH_REGEX = "master"

    @staticmethod
    def is_repo(path):
        """
        Detects if path is inside a valid Git repo.

        Parameters:
        path: Path

        Returns:
        True, path is in Git repo. False, path is not in Git repo.
        """
        head, tail = os.path.split(os.path.abspath(path))
        while tail:
            dot_git_path = os.path.join(head, tail, ".git")
            if os.path.exists(dot_git_path):
                return True
            path = head
            head, tail = os.path.split(path)
        return False

    def get_properties(self):
        tag, tag_version = self.get_latest_version_tag()
        return {
            "scm_change_id": "x" + self._cmd(CMD_SHORT_HASH),
            "commit_count": self._num_commits_since(tag),
            "tag_version": tag_version,
            "branch": self._branch(),
        }

    def _branch(self):
        refs = self._cmd(CMD_SHOW_REFS).split("\n")
        branch_candidate = None
        for ref in refs:
            if "detached" in ref:
                continue
            if ref.startswith("refs/heads/"):
                return ref[len("refs/heads/") :]

        if branch_candidate is None and "GIT_BRANCH" in os.environ:
            return os.environ["GIT_BRANCH"]

        return branch_candidate

    def _num_commits_since(self, tag):
        cmd = CMD_REV_COUNT.format(start=tag, end="HEAD")
        return self._cmd(cmd)

    def _get_first_commit(self):
        all_revistions = self._cmd(CMD_REV_LIST)
        return all_revistions.partition("\n")[0]

    def get_latest_version_tag(self):
        try:
            tag = self._cmd(CMD_LATEST_VERSION_TAG)
            commit = tag
        except GitCommandError:
            LOG.info("unable to find a valid version tag. using the default.")
            tag = "v" + DEFAULT_TAG_VERSION
            commit = self._get_first_commit()
        extracted_version = extract_tag_version(tag)
        return (commit, extracted_version)

    def _cmd(self, cmd):
        """
        Arguments:
            cmd - command to run

        Returns the output of the command.
        """
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            cwd=self._path,
        )
        stdout, stderr = proc.communicate()
        stdout = stdout.decode("utf-8").strip()
        stderr = stderr.decode("utf-8").strip()
        LOG.debug("stdout of cmd: " + stdout)
        LOG.debug("stderr of cmd: " + stderr)

        if proc.returncode == 0:
            return stdout
        raise GitCommandError(
            "'{0}' returned an error code {1}".format(cmd, proc.returncode)
        )



import sys
import pytest
from vcver.scm.git import Git


@pytest.mark.linux
@pytest.mark.parametrize(
    "inp, out",
    [
        ("/a/b", False),
        ("/a", False),
        ("/a/b/..", False),
        ("/x/", False),
        ("/..", False),
    ],
)
@pytest.mark.usefixtures("no_fs")
def test_linux_is_repo(inp, out):
    """Fails when non-repo case is not detected for linux."""
    assert Git.is_repo(inp) == out


@pytest.mark.win32
@pytest.mark.parametrize(
    "inp, out",
    [
        ("c:\\a\\b", False),
        ("c:\\a", False),
        ("c:\\x\\..", False),
        ("c:\\f\\", False),
        ("c:\\..", False),
    ],
)
@pytest.mark.usefixtures("no_fs")
def test_win32_is_repo(inp, out):
    """Fails when non-repo case is not detected for win32."""
    assert Git.is_repo(inp) == out



import vcver
import os
import pytest
from vcver.version import _get_scm


def test_happy_case():
    """
    instead of setting up a git repo,
    just testing that vcver's own
    version string matches the
    expected format.
    """
    v = vcver.get_version()
    version, scm = v.split("+")

    try:
        if "master" in scm:
            # major / minor should be an int
            major, version_remainder = _get_leading_version_part(version)
            int(major)
            minor, version_remainder = _get_leading_version_part(
                version_remainder[0]
            )
            int(minor)
            patch, version_remainder = _get_leading_version_part(
                version_remainder[0]
            )
            int(patch)
            dev, version_remainder = _get_leading_version_part(
                version_remainder[0]
            )
            # commitcount should be an int
            int(dev[3:])
            assert dev.startswith("dev")
        else:
            should_be_zero, dev_commit_count = version.split(".")
            assert should_be_zero == "0"
            assert dev_commit_count.startswith("dev")
    except Exception as ex:
        msg = "{0} Generated  Version: {1}".format(ex, version)
        raise Exception(msg)

    branch, change_id = scm.split(".")
    assert change_id.startswith("x")
    change_id = change_id[1:]
    # scm_change_id should be a git hash
    assert set(change_id) & set("abcdef1234567890") == set(change_id)


@pytest.mark.skip
def _get_leading_version_part(version):
    version_parts = version.split(".", 1)
    return version_parts[0], version_parts[1:]


def test_child_directory_detected_as_git_repo():
    """
    """
    assert _get_scm(None, os.path.join(os.curdir, "vcver")) is not None



import pytest
from vcver.version_string import make_string_pep440_compatible


@pytest.mark.parametrize(
    "inp, out",
    [
        ("master", "master"),
        ("develop", "develop"),
        ("hotfix", "hotfix"),
        ("feature/foo", "featurefoo"),
        ("feature_bar", "featurebar"),
        ("feature+baz", "featurebaz"),
        ("feature/ABC-123", "featureABC123"),
    ],
)
def test_make_string_pep440_compatible(inp, out):
    assert make_string_pep440_compatible(inp) == out



import logging
from .version import get_version

LOG = logging.getLogger(__name__)


def setup_keywords_entry_point(dist, attr, value):
    logging.basicConfig()
    version = get_version(**value)
    dist.metadata.version = version



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

base = os.path.dirname(os.path.abspath(__file__))

README_PATH = os.path.join(base, "README.rst")

install_requires = ["packaging"]

tests_require = []

setup(
    name="vcver",
    version="0.2.1",
    description="provide package versions with version control data.",
    long_description=open(README_PATH).read(),
    author="Yusuke Tsutsumi",
    author_email="yusuke@tsutsumi.io",
    url="https://github.com/toumorokoshi/vcver-python",
    packages=find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Software Distribution",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],
    tests_require=tests_require,
    entry_points={
        "distutils.setup_keywords": ["vcver=vcver:setup_keywords_entry_point"]
    },
)



import pytest
from vcver.scm.base import extract_tag_version


@pytest.mark.parametrize("inp, out", [("v0.1", "0.1"), ("oogabooga", None)])
def test_extract_version(inp, out):
    assert extract_tag_version(inp) == out
